# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Primary item-list abstraction.
"""

from __future__ import annotations

import warnings
from typing import overload

import numpy as np
import pandas as pd
import pyarrow as pa
import torch
from numpy.typing import ArrayLike, NDArray
from typing_extensions import (
    Any,
    Literal,
    LiteralString,
    Sequence,
    TypeAlias,
    cast,
)

from lenskit.diagnostics import DataWarning

from .checks import array_is_null, check_1d
from .mtarray import MTArray, MTGenericArray
from .types import IDArray, IDSequence
from .vocab import Vocabulary

Backend: TypeAlias = Literal["numpy", "torch", "arrow"]


class ItemList:
    """
    Representation of a (usually ordered) list of items, possibly with scores
    and other associated data; many components take and return item lists.  Item
    lists are to be treated as **immutable** — create a new list with modified
    data, do not do in-place modifications of the list itself or the arrays or
    data frame it returns.

    An item list logically a list of rows, each of which is an item with
    multiple fields.  A designated field, ``score``, is available through the
    :meth:`scores` method, and is always single-precision floating-point.

    Item lists can be subset as an array (e.g. ``items[selector]``), where
    integer indices (or arrays thereof), boolean arrays, and slices are allowed
    as selectors.

    When an item list is pickled, it is pickled compactly but only for CPUs: the
    vocabulary is dropped (after ensuring both IDs and numbers are computed),
    and all arrays are pickled as NumPy arrays.  This makes item lists compact
    to serialize and transmit, but does mean that that serializing an item list
    whose scores are still on the GPU will deserialize on the CPU in the
    receiving process.  This is usually not a problem, because item lists are
    typically used for small lists of items, not large data structures that need
    to remain in shared memory.

    .. note::

        Naming for fields and accessor methods is tricky, because the usual
        convention for a data frame is to use singular column names (e.g.
        “item_id”, “score”) instead of plural (“item_ids”, “scores”) — the data
        frame, like a database table, is a list of instances, and the column
        names are best interpreted as naming attributes of individual instances.

        However, when working with a list of e.g. item IDs, it is more natural —
        at least to this author — to use plural names: ``item_ids``.  Since this
        class is doing somewhat double-duty, representing a list of items along
        with associated data, as well as a data frame of columns representing
        items, the appropriate naming is not entirely clear.  The naming
        convention in this class is therefore as follows:

        * Field names are singular (``item_id``, ``score``).
        * Named accessor methods are plural (:meth:`item_ids`, :meth:`scores`).
        * Both singular and plural forms are accepted for item IDs numbers, and
          scores in the keyword arguments.  Other field names should be
          singular.

    .. todo::

        Right now, selection / subsetting only happens on the CPU, and will move
        data to the CPU for the subsetting operation.  There is no reason, in
        principle, why we cannot subset on GPU.  Future revisions may add
        support for this.

    Args:
        source:
            A source item list. If provided and an :class:`ItemList`, its fields
            and data are used to initialize any aspects of the item list that
            are not provided in the other arguments.  Otherwise, it is
            interpreted as ``item_ids``.
        item_ids:
            A list or array of item identifiers. ``item_id`` is accepted as an
            alternate name.
        item_nums:
            A list or array of item numbers. ``item_num`` is accepted as an
            alternate name.
        vocabulary:
            A vocabulary to translate between item IDs and numbers.
        ordered:
            Whether the list has a meaningful order.
        scores:
            An array of scores for the items.  Pass the value ``False`` to
            remove the scores when copying from a source list.
        fields:
            Additional fields, such as ``score`` or ``rating``.  Field names
            should generally be singular; the named keyword arguments and
            accessor methods are plural for readability (“get the list of item
            IDs”).  Pass the value ``False`` to remove the field when copying
            from a source list.
    """

    ordered: bool = False
    "Whether this list has a meaningful order."
    _len: int
    _ids: IDArray | None = None
    _numbers: MTArray[np.int32] | None = None
    _vocab: Vocabulary | None = None
    _ranks: MTArray[np.int32] | None = None
    _fields: dict[str, MTGenericArray]

    def __init__(
        self,
        source: ItemList | IDSequence | None = None,
        *,
        item_ids: IDSequence | None = None,
        item_nums: NDArray[np.int32] | pd.Series[int] | Sequence[int] | ArrayLike | None = None,
        vocabulary: Vocabulary | None = None,
        ordered: bool | None = None,
        scores: NDArray[np.generic]
        | torch.Tensor
        | ArrayLike
        | Literal[False]
        | np.floating
        | float
        | None = None,
        **fields: NDArray[np.generic] | torch.Tensor | ArrayLike | Literal[False],
    ):
        if isinstance(source, ItemList):
            self.__dict__.update(source.__dict__)
            eff_fields = source._fields | fields
        else:
            eff_fields = fields

        if ordered is not None:
            self.ordered = ordered
        elif source is None:
            self.ordered = False

        if vocabulary is not None:
            if not isinstance(vocabulary, Vocabulary):  # pragma: nocover
                raise TypeError(f"expected Vocabulary, got {type(vocabulary)}")
            self._vocab = vocabulary

        # handle aliases for item ID/number columns
        if item_ids is None and "item_id" in fields:
            item_ids = np.asarray(cast(Any, fields["item_id"]))
        if source is not None and not isinstance(source, ItemList):
            if item_ids is None:
                item_ids = source
                source = None
            else:
                raise ValueError("cannot specify both item_ids & item ID source")

        if item_nums is None and "item_num" in fields:
            item_nums = np.asarray(cast(Any, fields["item_num"]))
            if not issubclass(item_nums.dtype.type, np.integer):
                raise TypeError("item numbers not integers")

        # empty list
        if item_ids is None and item_nums is None and source is None:
            self._ids = np.ndarray(0, dtype=np.int32)
            self._numbers = MTArray(np.ndarray(0, dtype=np.int32))
            self._len = 0

        if item_ids is not None:
            if len(item_ids):
                self._ids = np.asarray(item_ids)
            else:
                self._ids = np.ndarray(0, dtype=np.int32)
            if not issubclass(self._ids.dtype.type, (np.integer, np.str_, np.bytes_, np.object_)):
                raise TypeError(
                    f"item IDs not integers, strings, or objects (type: {self._ids.dtype})"
                )

            check_1d(self._ids, label="item_ids")
            self._len = len(item_ids)
            # clear numbers if we got them from the source
            if source is not None and source._numbers is not None:
                del self._numbers

        if item_nums is not None:
            if not len(item_nums):  # type: ignore
                item_nums = np.ndarray(0, dtype=np.int32)
            self._numbers = MTArray(item_nums)
            check_1d(self._numbers, getattr(self, "_len", None), label="item_nums")
            self._len = self._numbers.shape[0]
            # clear IDs if we got them from the source
            if source is not None and source._ids is not None:
                del self._ids

        if scores is False:  # check 'is False' to distinguish from None
            scores = None
        else:
            if scores is None and "score" in eff_fields:
                scores = np.require(eff_fields["score"], dtype=np.float32)
            elif scores is not None:
                if "score" in fields:  # pragma: nocover
                    raise ValueError("cannot specify both scores= and score=")

                if np.isscalar(scores):
                    scores = np.full(self._len, scores, dtype=np.float32)
                else:
                    scores = np.require(scores, np.float32)

        if "rank" in fields:
            if ordered is False:
                warnings.warn(
                    "ranks provided but ordered=False, dropping ranks", DataWarning, stacklevel=2
                )
            else:
                self._ranks = check_1d(
                    MTArray(np.require(fields["rank"], np.int32)), self._len, label="ranks"
                )
                if self._len and self._ranks.numpy()[0] != 1:
                    warnings.warn("ranks do not begin with 1", DataWarning, stacklevel=2)
                self.ordered = True

        # convert fields and drop singular ID/number aliases
        self._fields = {}
        if not array_is_null(scores):
            assert scores is not None
            self._fields["score"] = check_1d(MTArray(scores), self._len, label="scores")

        for name, data in eff_fields.items():
            if (
                name not in ("item_id", "item_num", "score", "rank")
                and data is not False
                and not array_is_null(data)
            ):
                self._fields[name] = check_1d(MTArray(data), self._len, label=name)

    @classmethod
    def from_df(
        cls, df: pd.DataFrame, *, vocabulary: Vocabulary | None = None, keep_user: bool = False
    ) -> ItemList:
        """
        Create a item list from a Pandas data frame.  The frame should have
        ``item_num`` and/or ``item_id`` columns to identify the items; other
        columns (e.g. ``score`` or ``rating``) are added as fields. If the data
        frame has user columns (``user_id`` or ``user_num``), those are dropped
        by default.

        Args:
            df:
                The data frame to turn into an item list.
            vocabulary:
                The item vocabulary.
            keep_user:
                If ``True``, keeps user ID/number columns instead of dropping them.
        """
        ids = df["item_id"].values if "item_id" in df.columns else None
        nums = df["item_num"].values if "item_num" in df.columns else None
        if ids is None and nums is None:
            raise TypeError("data frame must have at least one of item_id, item_num columns")

        to_drop = ["item_id", "item_num"]
        if not keep_user:
            to_drop += ["user_id", "user_num"]

        fields = {f: df[f].values for f in df.columns if f not in to_drop}

        items = cls(
            item_ids=ids,  # type: ignore
            item_nums=nums,  # type: ignore
            vocabulary=vocabulary,
            **fields,  # type: ignore
        )

        return items

    @classmethod
    def from_arrow(
        cls,
        tbl: pa.StructArray | pa.ChunkedArray | pa.Table,
        *,
        vocabulary: Vocabulary | None = None,
    ) -> ItemList:
        """
        Create a item list from a Pandas table or structured array.  The table
        should have ``item_num`` and/or ``item_id`` columns to identify the
        items; other columns (e.g. ``score`` or ``rating``) are added as fields.
        If the data frame has user columns (``user_id`` or ``user_num``), those
        are dropped by default.

        Args:
            tbl:
                The Arrow table or array to convert to an item list.
            vocabulary:
                The item vocabulary.
        """
        if isinstance(tbl, pa.Table):
            tbl = tbl.to_struct_array()

        if isinstance(tbl, pa.ChunkedArray):
            tbl = tbl.combine_chunks()  # type: ignore
        assert isinstance(tbl, pa.StructArray)
        assert isinstance(tbl.type, pa.StructType)

        if hasattr(tbl.type, "names"):
            names = tbl.type.names  # type: ignore
        else:
            names = [tbl.type.field(i).name for i in range(tbl.type.num_fields)]

        ids = tbl.field("item_id") if "item_id" in names else None
        nums = tbl.field("item_num") if "item_num" in names else None
        if ids is None and nums is None:
            raise TypeError("data table must have at least one of item_id, item_num columns")

        to_drop = ["item_id", "item_num"]

        fields = {c: tbl.field(c).to_numpy() for c in names if c not in to_drop}
        items = cls(
            item_ids=ids,  # type: ignore
            item_nums=nums,  # type: ignore
            vocabulary=vocabulary,
            **fields,  # type: ignore
        )
        assert len(items) == len(tbl)
        return items

    @classmethod
    def from_vocabulary(cls, vocab: Vocabulary) -> ItemList:
        return ItemList(item_ids=vocab.ids(), item_nums=np.arange(len(vocab)), vocabulary=vocab)

    def clone(self) -> ItemList:
        """
        Make a shallow copy of the item list.
        """
        return ItemList(
            item_ids=self._ids,
            item_nums=self._numbers,
            vocabulary=self._vocab,
            ordered=self.ordered,
            **self._fields,  # type: ignore
        )

    @property
    def vocabulary(self) -> Vocabulary | None:
        "Get the item list's vocabulary, if available."
        return self._vocab

    def ids(self) -> IDArray:
        """
        Get the item IDs.

        Returns:
            An array of item identifiers.

        Raises:
            RuntimeError: if the item list was not created with IDs or a :class:`Vocabulary`.
        """
        if self._ids is None:
            if self._vocab is None:
                raise RuntimeError("item IDs not available (no IDs or vocabulary provided)")
            assert self._numbers is not None
            self._ids = self._vocab.ids(self._numbers.numpy())

        return self._ids

    @overload
    def numbers(
        self,
        format: Literal["numpy"] = "numpy",
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> NDArray[np.int32]: ...
    @overload
    def numbers(
        self,
        format: Literal["torch"],
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> torch.Tensor: ...
    @overload
    def numbers(
        self,
        format: Literal["arrow"],
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> pa.Array[pa.Int32Scalar]: ...
    @overload
    def numbers(
        self,
        format: LiteralString = "numpy",
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> ArrayLike: ...
    def numbers(
        self,
        format: LiteralString = "numpy",
        *,
        vocabulary: Vocabulary | None = None,
        missing: Literal["error", "negative"] = "error",
    ) -> ArrayLike:
        """
        Get the item numbers.

        Args:
            format:
                The array format to use.
            vocabulary:
                A alternate vocabulary for mapping IDs to numbers.  If provided,
                then the item list must have IDs (either stored, or through a
                vocabulary).

        Returns:
            An array of item numbers.

        Raises:
            RuntimeError: if the item list was not created with numbers or a
            :class:`Vocabulary`.
        """
        if vocabulary is not None and not vocabulary.compatible_with_numbers_from(self._vocab):
            # we need to translate vocabulary
            ids = self.ids()
            return vocabulary.numbers(ids, missing=missing)

        if self._numbers is None:
            if self._vocab is None:
                raise RuntimeError("item numbers not available (no IDs or vocabulary provided)")
            assert self._ids is not None
            self._numbers = MTArray(self._vocab.numbers(self._ids, missing="negative"))

        if missing == "error" and np.any(self._numbers.to("numpy") < 0):
            raise KeyError("item IDs")
        return self._numbers.to(format)

    @overload
    def scores(self, format: Literal["numpy"] = "numpy") -> NDArray[np.float32] | None: ...
    @overload
    def scores(self, format: Literal["torch"]) -> torch.Tensor | None: ...
    @overload
    def scores(self, format: Literal["arrow"]) -> pa.Array[pa.FloatScalar] | None: ...
    @overload
    def scores(
        self, format: Literal["pandas"], *, index: Literal["ids", "numbers"] | None = None
    ) -> pd.Series | None: ...
    @overload
    def scores(self, format: LiteralString = "numpy") -> ArrayLike | None: ...
    def scores(self, format: LiteralString = "numpy", **kwargs) -> ArrayLike | None:
        """
        Get the item scores (if available).
        """
        return self.field("score", format, **kwargs)

    @overload
    def ranks(self, format: Literal["numpy"] = "numpy") -> NDArray[np.int32] | None: ...
    @overload
    def ranks(self, format: Literal["torch"]) -> torch.Tensor | None: ...
    @overload
    def ranks(self, format: Literal["arrow"]) -> pa.Array[pa.Int32Scalar] | None: ...
    @overload
    def ranks(self, format: LiteralString = "numpy") -> ArrayLike | None: ...
    def ranks(self, format: LiteralString = "numpy") -> ArrayLike | None:
        """
        Get an array of ranks for the items in this list, if it is ordered.
        Unordered lists have no ranks.  The ranks are based on the order in the
        list, **not** on the score.

        Item ranks start with **1**, for compatibility with common practice in
        mathematically defining information retrieval metrics and operations.

        Returns:
            An array of item ranks, or ``None`` if the list is unordered.
        """
        if not self.ordered:
            return None

        if self._ranks is None:
            self._ranks = MTArray(np.arange(1, self._len + 1, dtype=np.int32))

        return self._ranks.to(format)

    @overload
    def field(
        self, name: str, format: Literal["numpy"] = "numpy"
    ) -> NDArray[np.floating] | None: ...
    @overload
    def field(self, name: str, format: Literal["torch"]) -> torch.Tensor | None: ...
    @overload
    def field(self, name: str, format: Literal["arrow"]) -> pa.Array | pa.Tensor | None: ...
    @overload
    def field(
        self,
        name: str,
        format: Literal["pandas"],
        *,
        index: Literal["ids", "numbers"] | None = None,
    ) -> pd.Series | None: ...
    @overload
    def field(self, name: str, format: LiteralString) -> ArrayLike | None: ...
    def field(
        self, name: str, format: LiteralString = "numpy", *, index: LiteralString | None = None
    ) -> object:
        val = self._fields.get(name, None)
        if val is None:
            return None
        elif format == "pandas":
            idx = None
            vs = val.to("numpy")
            if index == "ids":
                idx = pd.Index(self.ids(), name="item_id")
            elif index == "numbers":
                idx = pd.Index(self.numbers(), name="item_num")
            elif index:  # pragma: nocover
                raise ValueError(f"unsupported Pandas index {index}")
            return pd.Series(vs, index=idx)
        else:
            return val.to(format)

    def to_df(self, *, ids: bool = True, numbers: bool = True) -> pd.DataFrame:
        """
        Convert this item list to a Pandas data frame.  It has the following columns:

        * ``item_id`` — the item IDs (if available and ``ids=True``)
        * ``item_num`` — the item numbers (if available and ``numbers=True``)
        * ``score`` — the item scores
        * ``rank`` — the item ranks (if the list is ordered)
        * all other defined fields, using their field names
        """
        cols = {}
        if ids and (self._ids is not None or self._vocab is not None):
            cols["item_id"] = self.ids()
        if numbers and (self._numbers is not None or self._vocab is not None):
            cols["item_num"] = self.numbers()
        # we need to have numbers or ids, or it makes no sense
        if "item_id" not in cols and "item_num" not in cols:
            if ids and not numbers:
                raise RuntimeError("item list has no vocabulary, cannot compute IDs")
            elif numbers and not ids:
                raise RuntimeError("item list has no vocabulary, cannot compute numbers")
            else:
                raise RuntimeError("cannot create item data frame without identifiers or numbers")

        if "score" in self._fields:
            cols["score"] = self.scores()
        if self.ordered:
            cols["rank"] = self.ranks()
        # add remaining fields
        cols.update((k, v.numpy()) for (k, v) in self._fields.items() if k != "score")
        return pd.DataFrame(cols)

    @overload
    def to_arrow(
        self,
        *,
        ids: bool = True,
        numbers: bool = False,
        type: Literal["table"] = "table",
        columns: dict[str, pa.DataType] | None = None,
    ) -> pa.Table: ...
    @overload
    def to_arrow(
        self,
        *,
        ids: bool = True,
        numbers: bool = False,
        type: Literal["array"],
        columns: dict[str, pa.DataType] | None = None,
    ) -> pa.StructArray: ...
    def to_arrow(
        self,
        *,
        ids: bool = True,
        numbers: bool = False,
        type: Literal["table", "array"] = "table",
        columns: dict[str, pa.DataType] | None = None,
    ):
        """
        Convert the item list to a Pandas table.
        """
        arrays = []
        names = []
        if columns is not None and len(self) == 0:
            arrays = [pa.array([], ft) for ft in columns.values()]
            names = list(columns.keys())
        else:
            if columns is None:
                columns = self.arrow_types(ids=ids, numbers=numbers)

            for c_name, c_type in columns.items():
                names.append(c_name)
                if c_name == "item_id":
                    arrays.append(pa.array(self.ids()))
                elif c_name == "item_num":
                    arrays.append(self.numbers("arrow"))
                elif c_name == "rank":
                    if self.ordered:
                        arrays.append(self.ranks("arrow"))
                    else:
                        warnings.warn("requested rank column for unordered list", DataWarning)
                        arrays.append(pa.nulls(len(self), c_type))
                elif fld := self._fields.get(c_name, None):
                    arrays.append(fld.arrow())
                else:
                    warnings.warn(f"unknown field {c_name}", DataWarning)

        if type == "table":
            return pa.Table.from_arrays(arrays, names)
        elif type == "array":
            return pa.StructArray.from_arrays(arrays, names)
        else:  # pragma: nocover
            raise ValueError(f"unsupported target type {type}")

    def arrow_types(self, *, ids: bool = True, numbers: bool = False) -> dict[str, pa.DataType]:
        """
        Get the Arrow data types for this item list.
        """
        types: dict[str, pa.DataType] = {}
        if len(self) == 0:
            return types

        if ids:
            if self._ids is not None:
                types["item_id"] = _arrow_type(self._ids.dtype)
            elif self._vocab is not None:
                types["item_id"] = _arrow_type(self._vocab.ids().dtype)

        if numbers and (self._numbers is not None or self._vocab is not None):
            types["item_num"] = pa.int32()

        if self.ordered:
            types["rank"] = pa.int32()

        for name, f in self._fields.items():
            types[name] = _arrow_type(f.numpy().dtype)

        return types

    def __len__(self):
        return self._len

    def __getitem__(
        self,
        sel: NDArray[np.bool_] | NDArray[np.integer] | Sequence[int] | torch.Tensor | int | slice,
    ) -> ItemList:
        """
        Subset the item list.

        .. todo::
            Support on-device masking.

        Args:
            sel:
                The items to select. Can be either a Boolean array of the same
                length as the list that is ``True`` to indicate selected items,
                or an array of indices of the items to retain (in order in the
                list, starting from 0).
        """
        if np.isscalar(sel):
            sel = np.array([sel])
        elif not isinstance(sel, slice):
            sel = np.asarray(sel)

        # sel is now a selection array, or it is a slice. numpy supports both.
        iids = self._ids[sel] if self._ids is not None else None
        nums = self._numbers.numpy()[sel] if self._numbers is not None else None
        flds = {n: f.numpy()[sel] for (n, f) in self._fields.items()}
        return ItemList(
            item_ids=iids,
            item_nums=nums,
            vocabulary=self._vocab,
            ordered=self.ordered,
            **flds,  # type: ignore
        )

    def __getstate__(self) -> dict[str, object]:
        state: dict[str, object] = {"ordered": self.ordered, "len": self._len}
        if self._ids is not None:
            state["ids"] = self._ids
        elif self._vocab is not None:
            # compute the IDs so we can save them
            state["ids"] = self.ids()

        if self._numbers is not None:
            state["numbers"] = self._numbers.numpy()
        elif self._vocab is not None:
            state["numbers"] = self.numbers()

        state.update(("field_" + k, v.numpy()) for (k, v) in self._fields.items())
        return state

    def __setstate__(self, state: dict[str, Any]):
        self.ordered = state["ordered"]
        self._len = state["len"]
        self._ids = state.get("ids", None)
        if "numbers" in state:
            self._numbers = MTArray(state["numbers"])
        self._fields = {k[6:]: MTArray(v) for (k, v) in state.items() if k.startswith("field_")}

    def __str__(self) -> str:
        return f"<ItemList of {self._len} items>"

    def __repr__(self) -> str:
        fnames = ", ".join(self._fields.keys())
        nf = len(self._fields)
        return f"<ItemList of {self._len} items with {nf} fields ({fnames})>"


def _arrow_type(dtype: np.dtype) -> pa.DataType:
    if dtype == np.object_:
        return pa.utf8()
    else:
        return pa.from_numpy_dtype(dtype)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
User-based k-NN collaborative filtering.
"""

# pyright: basic
from __future__ import annotations

import warnings

import numpy as np
import pandas as pd
import structlog
import torch
from scipy.sparse import csr_array
from typing_extensions import NamedTuple, Optional, Self, override

from lenskit import util
from lenskit.data import Dataset, FeedbackType, ItemList, QueryInput, RecQuery
from lenskit.data.vocab import Vocabulary
from lenskit.diagnostics import DataWarning
from lenskit.math.sparse import normalize_sparse_rows, safe_spmv, torch_sparse_to_scipy
from lenskit.parallel.config import ensure_parallel_init
from lenskit.pipeline import Component, Trainable

_log = structlog.stdlib.get_logger(__name__)


class UserKNNScorer(Component, Trainable):
    """
    User-user nearest-neighbor collaborative filtering with ratings. This
    user-user implementation is not terribly configurable; it hard-codes design
    decisions found to work well in the previous Java-based LensKit code.

    Args:
        nnbrs:
            the maximum number of neighbors for scoring each item (``None`` for
            unlimited).
        min_nbrs:
            The minimum number of neighbors for scoring each item.
        min_sim:
            Minimum similarity threshold for considering a neighbor.  Must be
            positive; if less than the smallest 32-bit normal (:math:`1.175
            \\times 10^{-38}`), is clamped to that value.
        feedback:
            Control how feedback should be interpreted.  Specifies defaults for
            the other settings, which can be overridden individually; can be one
            of the following values:

            ``explicit``
                Configure for explicit-feedback mode: use rating values, and
                predict using weighted averages.  This is the default setting.

            ``implicit``
                Configure for implicit-feedback mode: ignore rating values, and
                predict using the sums of similarities.
    """

    nnbrs: int
    min_nbrs: int
    min_sim: float
    feedback: FeedbackType

    users_: Vocabulary
    "The index of user IDs."
    items_: Vocabulary
    "The index of item IDs."
    user_means_: torch.Tensor | None
    "Mean rating for each known user."
    user_vectors_: torch.Tensor
    "Normalized rating matrix (CSR) to find neighbors at prediction time."
    user_ratings_: csr_array
    "Centered but un-normalized rating matrix (COO) to find neighbor ratings."

    def __init__(
        self,
        nnbrs: int,
        min_nbrs: int = 1,
        min_sim: float = 1.0e-6,
        feedback: FeedbackType = "explicit",
    ):
        self.nnbrs = nnbrs
        self.min_nbrs = min_nbrs
        if min_sim < 0:
            raise ValueError("minimum similarity must be positive")
        elif min_sim == 0:
            f4i = np.finfo("f4")
            self.min_sim = float(f4i.smallest_normal)
            _log.warning("minimum similarity %e is too low, using %e", min_sim, self.min_sim)
        else:
            self.min_sim = min_sim

        self.feedback = feedback

    @property
    def is_trained(self) -> bool:
        return hasattr(self, "users_")

    @override
    def train(self, data: Dataset) -> Self:
        """
        "Train" a user-user CF model.  This memorizes the rating data in a format that is usable
        for future computations.

        Args:
            ratings(pandas.DataFrame): (user, item, rating) data for collaborative filtering.
        """
        ensure_parallel_init()
        rmat = data.interaction_matrix(
            "torch", field="rating" if self.feedback == "explicit" else None
        )
        assert rmat.is_sparse_csr

        if self.feedback == "explicit":
            rmat, means = normalize_sparse_rows(rmat, "center")
            if np.allclose(rmat.values(), 0.0):
                _log.warning("normalized ratings are zero, centering is not recommended")
                warnings.warn(
                    "Ratings seem to have the same value, centering is not recommended.",
                    DataWarning,
                )
        else:
            means = None

        normed, _norms = normalize_sparse_rows(rmat, "unit")

        self.user_vectors_ = normed
        self.user_ratings_ = torch_sparse_to_scipy(rmat).tocsr()
        self.users_ = data.users.copy()
        self.user_means_ = means
        self.items_ = data.items.copy()

        return self

    @override
    def __call__(self, query: QueryInput, items: ItemList) -> ItemList:
        """
        Compute predictions for a user and items.

        Args:
            user: the user ID
            items (array-like): the items to predict
            ratings (pandas.Series):
                the user's ratings (indexed by item id); if provided, will be used to
                recompute the user's bias at prediction time.

        Returns:
            pandas.Series: scores for the items, indexed by item id.
        """
        query = RecQuery.create(query)
        watch = util.Stopwatch()
        log = _log.bind(user_id=query.user_id, n_items=len(items))

        udata = self._get_user_data(query)
        if udata is None:
            log.debug("user has no ratings, skipping")
            return ItemList(items, scores=np.nan)

        uidx, ratings, umean = udata
        assert ratings.shape == (len(self.items_),)  # ratings is a dense vector

        # now ratings has vbeen normalized to be a mean-centered unit vector
        # this means we can dot product to score neighbors
        # score the neighbors!
        nbr_sims = safe_spmv(self.user_vectors_, ratings)
        assert nbr_sims.shape == (len(self.users_),)
        if uidx is not None:
            # zero out the self-similarity
            nbr_sims[uidx] = 0

        # get indices for these neighbors
        nbr_idxs = torch.arange(len(self.users_), dtype=torch.int64)

        nbr_mask = nbr_sims >= self.min_sim

        kn_sims = nbr_sims[nbr_mask]
        kn_idxs = nbr_idxs[nbr_mask]
        if len(kn_sims) > 0:
            log.debug(
                "found %d candidate neighbors (of %d total), max sim %0.4f",
                len(kn_sims),
                len(self.users_),
                torch.max(kn_sims).item(),
            )
        else:
            log.warning("no candidate neighbors found")
            return ItemList(items, scores=np.nan)

        assert not torch.any(torch.isnan(kn_sims))

        iidxs = items.numbers(vocabulary=self.items_, missing="negative")
        iidxs = torch.from_numpy(iidxs).to(torch.int64)

        ki_mask = iidxs >= 0
        usable_iidxs = iidxs[ki_mask]

        scores = score_items_with_neighbors(
            log,
            usable_iidxs,
            kn_idxs,
            kn_sims,
            self.user_ratings_,
            self.nnbrs,
            self.min_nbrs,
            self.feedback == "explicit",
        )

        scores += umean

        results = pd.Series(scores, index=items.ids()[ki_mask.numpy()], name="prediction")
        results = results.reindex(items.ids())

        log.debug(
            "scored %d items in %s",
            results.notna().sum(),
            watch,
        )
        return ItemList(items, scores=results.values)  # type: ignore

    def _get_user_data(self, query: RecQuery) -> Optional[UserRatings]:
        "Get a user's data for user-user CF"

        index = self.users_.number(query.user_id, missing=None)

        if query.user_items is None:
            if index is None:
                _log.warning("user %s has no ratings and none provided", query.user_id)
                return None

            assert index >= 0
            row = self.user_vectors_[index].to_dense()
            if self.feedback == "explicit":
                assert self.user_means_ is not None
                umean = self.user_means_[index].item()
            else:
                umean = 0
            return UserRatings(index, row, umean)
        else:
            _log.debug("using provided item history")
            ratings = torch.zeros(len(self.items_), dtype=torch.float32)
            ui_nos = query.user_items.numbers("torch", missing="negative", vocabulary=self.items_)
            ui_mask = ui_nos >= 0

            if self.feedback == "explicit":
                urv = query.user_items.field("rating", "torch")
                if urv is None:
                    _log.warning("user %s has items but no ratings", query.user_id)
                    return None

                umean = urv.mean().item()
                ratings[ui_nos[ui_mask]] = urv[ui_mask] - umean
            else:
                umean = 0
                ratings[ui_nos[ui_mask]] = 1.0

            return UserRatings(index, ratings, umean)

    def __str__(self):
        return "UserUser(nnbrs={}, min_sim={})".format(self.nnbrs, self.min_sim)


class UserRatings(NamedTuple):
    """
    Dense user ratings.
    """

    index: int | None
    ratings: torch.Tensor
    mean: float


def score_items_with_neighbors(
    log: structlog.stdlib.BoundLogger,
    items: torch.Tensor,
    nbr_rows: torch.Tensor,
    nbr_sims: torch.Tensor,
    ratings: csr_array,
    max_nbrs: int,
    min_nbrs: int,
    average: bool,
) -> np.ndarray[tuple[int], np.dtype[np.float64]]:
    # select a sub-matrix for further manipulation
    (ni,) = items.shape
    (nrow, ncol) = ratings.shape
    # do matrix surgery
    nbr_rates = ratings[nbr_rows.numpy(), :]
    nbr_rates = nbr_rates[:, items.numpy()]

    nbr_t = nbr_rates.transpose().tocsr()

    # count nbrs for each item
    counts = np.diff(nbr_t.indptr)
    assert counts.shape == items.shape

    log.debug(
        "scoring items",
        max_count=np.max(counts),
        nbr_shape=nbr_rates.shape,
    )

    # fast-path items with small neighborhoods
    fp_mask = counts <= max_nbrs
    results = np.full(ni, np.nan)
    nbr_fp = nbr_rates[:, fp_mask]
    results[fp_mask] = nbr_fp.T @ nbr_sims

    if average:
        nbr_fp_ones = csr_array((np.ones(nbr_fp.nnz), nbr_fp.indices, nbr_fp.indptr), nbr_fp.shape)
        tot_sims = nbr_fp_ones.T @ nbr_sims
        assert np.all(np.isfinite(tot_sims))
        results[fp_mask] /= tot_sims

    # clear out too-small neighborhoods
    results[counts < min_nbrs] = torch.nan

    # deal with too-large items
    exc_mask = counts > max_nbrs
    n_bad = np.sum(exc_mask)
    if n_bad:
        log.debug("scoring %d slow-path items", n_bad)

    bads = np.argwhere(exc_mask)[:, 0]
    for badi in bads:
        s, e = nbr_t.indptr[badi : (badi + 2)]

        bi_users = nbr_t.indices[s:e]
        bi_rates = torch.from_numpy(nbr_t.data[s:e])
        bi_sims = nbr_sims[bi_users]

        tk_vs, tk_is = torch.topk(bi_sims, max_nbrs)
        sum = torch.sum(tk_vs)
        if average:
            results[badi] = torch.dot(tk_vs, bi_rates[tk_is]) / sum
        else:
            results[badi] = sum

    return results



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT
"""
Types used across LensKit.
"""

# pyright: strict
from __future__ import annotations

from typing import Any, Sequence, TypeAlias

import numpy as np

SeedLike: TypeAlias = int | np.integer[Any] | Sequence[int] | np.random.SeedSequence
"""
Type for RNG seeds (see `SPEC 7`_).

.. _SPEC 7: https://scientific-python.org/specs/spec-0007/
"""

RNGLike: TypeAlias = np.random.Generator | np.random.BitGenerator
"""
Type for random number generators as inputs (see `SPEC 7`_).

.. _SPEC 7: https://scientific-python.org/specs/spec-0007/
"""

RNGInput: TypeAlias = SeedLike | RNGLike | None
"""
Type for RNG inputs (see `SPEC 7`_).

.. _SPEC 7: https://scientific-python.org/specs/spec-0007/
"""



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Miscellaneous utility functions.
"""

import logging
from textwrap import dedent
from typing import Any, Protocol, TypeVar, runtime_checkable

from .random import derivable_rng, random_generator, set_global_rng
from .timing import Stopwatch

try:
    import resource
except ImportError:
    resource = None

_log = logging.getLogger(__name__)

__all__ = [
    "Stopwatch",
    "derivable_rng",
    "random_generator",
    "set_global_rng",
    "clean_str",
]

A = TypeVar("A")


@runtime_checkable
class ParamContainer(Protocol):
    def get_params(self, deep=True) -> dict[str, Any]: ...


def max_memory():
    "Get the maximum memory use for this process"
    if resource:
        res = resource.getrusage(resource.RUSAGE_SELF)
        return "%.1f MiB" % (res.ru_maxrss / 1024,)
    else:
        return "unknown"


def cur_memory():
    "Get the current memory use for this process"
    if resource:
        res = resource.getrusage(resource.RUSAGE_SELF)
        return "%.1f MiB" % (res.ru_idrss,)
    else:
        return "unknown"


def clean_str(s):
    return dedent(s).strip()



"""
Support for logging from worker processes.

This is internal support code, clients and LensKit implementers can usually
ignore it.
"""

from __future__ import annotations

import copy
import logging
import multiprocessing as mp
import pickle
import warnings
from dataclasses import dataclass
from hashlib import blake2b
from logging import Handler, LogRecord, getLogger
from threading import Lock

import structlog
import zmq
from structlog.typing import EventDict

from .config import CORE_PROCESSORS, active_logging_config, log_warning
from .monitor import get_monitor
from .processors import add_process_info
from .tasks import Task
from .tracing import lenskit_filtering_logger

_active_context: WorkerContext | None = None
_log = structlog.stdlib.get_logger(__name__)


@dataclass
class WorkerLogConfig:
    """
    Configuration for worker logging.
    """

    address: str
    level: int
    authkey: bytes | None = None

    @classmethod
    def current(cls):
        """
        Get the current worker logging configuration.
        """

        if _active_context is not None:
            return _active_context.config
        else:
            mon = get_monitor()
            cfg = active_logging_config()
            level = cfg.effective_level if cfg is not None else logging.INFO
            return cls(
                address=mon.log_address, level=level, authkey=bytes(mp.current_process().authkey)
            )


class WorkerContext:
    """
    Activate (and deactivate) a worker context.  This handles setup and teardown
    of logging, etc.

    Only one worker context can be active, regardless of how many threads are active.
    """

    config: WorkerLogConfig
    zmq: zmq.Context[zmq.Socket[bytes]]
    _log_handler: ZMQLogHandler

    def __init__(self, config: WorkerLogConfig):
        self.config = config
        if self.config.authkey is None:
            self.config.authkey = mp.current_process().authkey

    def start(self):
        global _active_context
        if _active_context is not None:
            raise RuntimeError("worker context already active")

        self.zmq = zmq.Context()
        self._log_handler = ZMQLogHandler(self.zmq, self.config)

        root = getLogger()
        root.addHandler(self._log_handler)
        root.setLevel(self.config.level)

        structlog.configure(
            [add_process_info]
            + CORE_PROCESSORS
            + [structlog.processors.ExceptionPrettyPrinter(), self._log_handler.send_structlog],
            wrapper_class=lenskit_filtering_logger(self.config.level),
            logger_factory=structlog.stdlib.LoggerFactory(),
            cache_logger_on_first_use=False,
        )
        warnings.showwarning = log_warning
        _log.debug("log context activated")

    def shutdown(self):
        root = getLogger()
        root.removeHandler(self._log_handler)

        self._log_handler.shutdown()
        self.zmq.term()

    def send_task(self, task: Task):
        self._log_handler.send_task(task)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.shutdown()


class ZMQLogHandler(Handler):
    _lock: Lock
    socket: zmq.Socket[bytes]
    key: bytes
    _render = structlog.processors.JSONRenderer()

    def __init__(self, zmq_context: zmq.Context, config: WorkerLogConfig):
        super().__init__()
        self.config = config
        self._lock = Lock()
        self.socket = zmq_context.socket(zmq.PUSH)
        self.socket.connect(config.address)

    def handle(self, record: LogRecord) -> LogRecord | bool:  # type: ignore
        # copy so other handlers don't have a problem
        record = copy.copy(record)

        # update messages for copyability
        if not hasattr(record, "message"):
            record.message = record.msg % record.args

        record.exc_info = None
        record.exc_text = None
        record.stack_info = None

        self._send_message(
            b"stdlib", record.name.encode(), pickle.dumps(record, pickle.HIGHEST_PROTOCOL)
        )

        return record

    def shutdown(self):
        self.socket.close()

    def send_structlog(self, logger, method, event_dict: EventDict):
        x = self._render(logger, method, {"method": method, "event": event_dict})
        if isinstance(x, str):
            x = x.encode()
        self._send_message(b"structlog", logger.name.encode(), x)

        raise structlog.DropEvent()

    def send_task(self, task: Task):
        _log.debug("sending updated task", task_id=task.task_id)
        self._send_message(
            b"lenskit.logging.tasks", str(task.task_id).encode(), task.model_dump_json().encode()
        )

    def _send_message(self, engine: bytes, name: bytes, data: bytes):
        key = self.config.authkey
        assert key is not None
        mb = blake2b(key=key)
        mb.update(engine)
        mb.update(name)
        mb.update(data)

        with self._lock:
            self.socket.send_multipart([engine, name, data, mb.digest()])



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Batch-run recommendation pipelines for evaluation.
"""

from __future__ import annotations

from ._predict import predict
from ._recommend import recommend
from ._results import BatchResults
from ._runner import BatchPipelineRunner, InvocationSpec

__all__ = ["BatchPipelineRunner", "BatchResults", "InvocationSpec", "predict", "recommend"]



from __future__ import annotations

from typing import Callable

from tqdm import tqdm

from ._base import Progress

__all__ = ["TQDMProgress", "tqdm"]


class TQDMProgress(Progress):
    """
    TQDM-based progress reporting. Useful for notebooks.
    """

    tqdm: tqdm
    _field_format: str | None = None

    def __init__(
        self,
        cls: Callable[..., tqdm],
        label: str | None,
        total: int | None,
        fields: dict[str, str | None],
    ):
        self.tqdm = tqdm(desc=label, total=total, leave=False)
        if fields:
            self._field_format = ", ".join(
                [f"{name}: {fs or None}" for (name, fs) in fields.items()]
            )

    def update(self, advance: int = 1, **kwargs: float | int | str):
        """
        Update the progress bar.
        """
        self.tqdm.update(advance)
        if self._field_format:
            self.tqdm.set_postfix_str(self._field_format.format(kwargs))

    def finish(self):
        """
        Finish and clean up this progress bar.  If the progresss bar is used as
        a context manager, this is automatically called on context exit.
        """
        self.tqdm.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.finish()



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

"""
Code to import MovieLens data sets into LensKit.
"""

import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TypeAlias
from zipfile import ZipFile

import numpy as np
import pandas as pd
import structlog

from .convert import from_interactions_df
from .dataset import Dataset

_log = structlog.stdlib.get_logger(__name__)

LOC: TypeAlias = Path | tuple[ZipFile, str]


class MLVersion(Enum):
    ML_100K = "ml-100k"
    ML_1M = "ml-1m"
    ML_10M = "ml-10m"
    ML_20M = "ml-20m"
    ML_25M = "ml-25m"
    ML_32M = "ml-32m"
    ML_LATEST_SMALL = "ml-latest-small"
    ML_LATEST = "ml-latest"
    ML_MODERN = "ml-modern"


@dataclass
class MLData:
    version: MLVersion
    source: Path | ZipFile
    prefix: str = ""

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if isinstance(self.source, ZipFile):
            self.source.close()

    def open_file(self, name: str):
        if isinstance(self.source, Path):
            return open(self.source / (self.prefix + name), "r")
        else:
            return self.source.open(self.prefix + name)


def load_movielens(path: str | Path) -> Dataset:
    """
    Load a MovieLens dataset.  The appropriate MovieLens format is detected
    based on the file contents.

    Args:
        path:
            The path to the dataset, either as an unpacked directory or a zip
            file.

    Returns:
        The dataset.
    """
    df = load_movielens_df(path)
    return from_interactions_df(df)


def load_movielens_df(path: str | Path) -> pd.DataFrame:
    """
    Load the ratings from a MovieLens dataset as a raw data frame.  The
    appropriate MovieLens format is detected based on the file contents.

    Args:
        path:
            The path to the dataset, either as an unpacked directory or a zip
            file.

    Returns:
        The ratings, with columns ``user``, ``item``, ``rating``, and
        ``timestamp``.
    """
    with _ml_detect_and_open(path) as ml:
        match ml.version:
            case MLVersion.ML_100K:
                return _load_ml_100k(ml)
            case MLVersion.ML_1M | MLVersion.ML_10M:
                return _load_ml_million(ml)
            case _:
                return _load_ml_modern(ml)


def _ml_detect_and_open(path: str | Path) -> MLData:
    loc = Path(path)
    ds: MLVersion

    if loc.is_file() and loc.suffix == ".zip":
        log = _log.bind(zipfile=str(loc))
        log.debug("opening zip file")
        zf = ZipFile(loc, "r")
        try:
            infos = zf.infolist()
            first = infos[0]
            if not first.is_dir:
                log.error("first entry is not directory")
                raise RuntimeError("invalid ML zip file")

            log.debug("base dir filename %s", first.filename)
            dsm = re.match(r"^(ml-(?:\d+[MmKk]|latest|latest-small))", first.filename)
            if not dsm:
                log.error("invalid directory name %s", first.filename)
                raise RuntimeError("invalid ML zip file")

            ds = MLVersion(dsm.group(1).lower())
            log.debug("found ML data set %s", ds)
            return MLData(ds, zf, first.filename)
        except Exception as e:  # pragma nocover
            zf.close()
            raise e
    else:
        log = _log.bind(dir=str(loc))
        log.debug("loading from directory")
        dsm = re.match(r"^(ml-\d+[MmKk])", loc.name)
        if dsm:
            ds = MLVersion(dsm.group(1))
            _log.debug("inferred data set %s from dir name", ds)
        else:
            _log.debug("checking contents for data type")
            if (loc / "u.data").exists():
                _log.debug("found u.data, interpreting as 100K")
                ds = MLVersion.ML_100K
            elif (loc / "ratings.dat").exists():
                if (loc / "tags.dat").exists():
                    _log.debug("found ratings.dat and tags.dat, interpreting as 10M")
                    ds = MLVersion.ML_10M
                else:
                    _log.debug("found ratings.dat but no tags, interpreting as 1M")
                    ds = MLVersion.ML_1M
            elif (loc / "ratings.csv").exists():
                _log.debug("found ratings.csv, interpreting as modern (20M and later)")
                ds = MLVersion.ML_MODERN
            else:
                _log.error("could not detect MovieLens data")
                raise RuntimeError("invalid ML directory")

        return MLData(ds, loc)


def _load_ml_100k(ml: MLData) -> pd.DataFrame:
    with ml.open_file("u.data") as data:
        return pd.read_csv(
            data,
            sep="\t",
            header=None,
            names=["user_id", "item_id", "rating", "timestamp"],
            dtype={
                "user_id": np.int32,
                "item_id": np.int32,
                "rating": np.float32,
                "timestamp": np.int32,
            },
        )


def _load_ml_million(ml: MLData) -> pd.DataFrame:
    with ml.open_file("ratings.dat") as data:
        return pd.read_csv(
            data,
            sep=":",
            header=None,
            names=["user_id", "_ui", "item_id", "_ir", "rating", "_rt", "timestamp"],
            usecols=[0, 2, 4, 6],
            dtype={
                "user_id": np.int32,
                "item_id": np.int32,
                "rating": np.float32,
                "timestamp": np.int32,
            },
        )


def _load_ml_modern(ml: MLData) -> pd.DataFrame:
    with ml.open_file("ratings.csv") as data:
        return pd.read_csv(
            data,
            dtype={
                "userId": np.int32,
                "movieId": np.int32,
                "rating": np.float32,
                "timestamp": np.int64,
            },
        ).rename(columns={"userId": "user_id", "movieId": "item_id"})



"""
Extended logger providing TRACE support.
"""

from __future__ import annotations

import logging
from typing import Any

import structlog


class TracingLogger(structlog.stdlib.BoundLogger):
    """
    Class for LensKit loggers with trace-level logging support.
    """

    def bind(self, **new_values: Any) -> TracingLogger:
        return super().bind(**new_values)  # type: ignore[return-value]

    def new(self, **new_values: Any) -> TracingLogger:
        return super().new(**new_values)  # type: ignore[return-value]

    def trace(self, event: str | None, *args: Any, **kw: Any):
        if args:
            kw["positional_args"] = args
        try:
            args, kwargs = self._process_event("trace", event, kw)
        except structlog.DropEvent:
            return None
        self._logger.debug(*args, **kwargs)


def lenskit_filtering_logger(level: int):
    """
    No-op filtering logger.
    """

    if level < logging.DEBUG:
        return TracingLogger

    def trace(self, event: str | None, *args: Any, **kw: Any):
        pass

    base = structlog.make_filtering_bound_logger(level)
    return type(f"LensKitLoggerFilter{level}", (base,), {"trace": trace})



import numpy as np

from lenskit.data import ItemList
from lenskit.data.query import QueryInput, RecQuery
from lenskit.pipeline import Component
from lenskit.util.random import DerivableSeed, RNGFactory, derivable_rng


class RandomSelector(Component):
    """
    Randomly select items from a candidate list.

    Args:
        n:
            The number of items to select, or ``-1`` to randomly permute the
            items.
        rng:
            The random number generator or specification (see :ref:`rng`).  This
            class supports derivable RNGs.
    """

    n: int
    rng: DerivableSeed
    _rng_factory: RNGFactory

    def __init__(self, n: int = -1, rng: DerivableSeed = None):
        self.n = n
        self.rng = rng
        self._rng_factory = derivable_rng(rng)

    def __call__(
        self, items: ItemList, query: QueryInput | None = None, n: int | None = None
    ) -> ItemList:
        """
        Args:
            items:
                The items from which to pick.
            query:
                The recommendation query; optional, and only consulted when the
                RNG seed is user-dependent.
            n:
                The number of items to select, overriding the configured value.
        """
        if n is None:
            n = self.n

        query = RecQuery.create(query)
        rng = self._rng_factory(query)

        if n < 0:
            n = len(items)
        else:
            n = min(n, len(items))

        if n > 0:
            picked = rng.choice(len(items), n, replace=False)
            return items[picked]
        else:
            return items[np.zeros(0, dtype=np.int32)]


class SoftmaxRanker(Component):
    """
    Stochastic top-N ranking with softmax sampling.

    This uses the “softmax” sampling method, a more efficient approximation of
    Plackett-Luce sampling than even the Gumbell trick, as documented by `Tim
    Vieira`_.  It expects a scored list of input items, and samples ``n`` items,
    with selection probabilities proportional to their scores.

    .. note::

        Negative scores are clamped to (approximately) zero.

    .. _`Tim Vieiera`: https://timvieira.github.io/blog/post/2019/09/16/algorithms-for-sampling-without-replacement/

    Args:
        n:
            The number of items to return (-1 to return unlimited).
        rng:
            The random number generator or specification (see :ref:`rng`).  This
            class supports derivable RNGs.
    """

    n: int
    rng: DerivableSeed
    _rng_factory: RNGFactory

    def __init__(self, n: int = -1, rng: DerivableSeed = None):
        self.n = n
        self.rng = rng
        self._rng_factory = derivable_rng(rng)

    def __call__(
        self, items: ItemList, query: QueryInput | None = None, n: int | None = None
    ) -> ItemList:
        query = RecQuery.create(query)
        rng = self._rng_factory(query)

        scores = items.scores()
        if scores is None:
            raise ValueError("item list must have scores")

        valid_mask = ~np.isnan(scores)
        valid_items = items[valid_mask]
        N = len(valid_items)
        if N == 0:
            return ItemList(item_ids=[], scores=[], ordered=True)

        if n is None or n < 0:
            n = self.n
        if n < 0 or n > N:
            n = N

        keys = -np.log(rng.uniform(0, 1, N))
        keys /= np.maximum(scores[valid_mask], 1.0e-10)

        picked = np.argsort(keys)[:n]
        return ItemList(valid_items[picked], ordered=True)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

from typing import Any, Callable

from typing_extensions import override

from .dataset import Dataset
from .items import ItemList
from .vocab import Vocabulary


class LazyDataset(Dataset):
    """
    A data set with an underlying load function, that doesn't call the function
    until data is actually needed.

    Args:
        loader:
            The function that will load the dataset when needed.
    """

    _delegate: Dataset | None = None
    _loader: Callable[[], Dataset]

    def __init__(self, loader: Callable[[], Dataset]):
        """
        Construct a lazy dataset.
        """
        self._loader = loader

    def delegate(self) -> Dataset:
        """
        Get the delegate data set, loading it if necessary.
        """
        if self._delegate is None:
            self._delegate = self._loader()
        return self._delegate

    @property
    @override
    def items(self) -> Vocabulary:
        return self.delegate().items

    @property
    @override
    def users(self) -> Vocabulary:
        return self.delegate().users

    @override
    def count(self, what: str) -> int:
        return self.delegate().count(what)

    @override
    def interaction_matrix(self, *args, **kwargs) -> Any:
        return self.delegate().interaction_matrix(*args, **kwargs)

    @override
    def interaction_log(self, *args, **kwargs) -> Any:
        return self.delegate().interaction_log(*args, **kwargs)

    @override
    def user_row(self, *args, **kwargs) -> ItemList | None:
        return self.delegate().user_row(*args, **kwargs)



# This file is part of LensKit.
# Copyright (C) 2018-2023 Boise State University
# Copyright (C) 2023-2024 Drexel University
# Licensed under the MIT license, see LICENSE.md for details.
# SPDX-License-Identifier: MIT

# pyright: basic
from __future__ import annotations

from concurrent.futures import ProcessPoolExecutor
from multiprocessing.context import SpawnContext, SpawnProcess
from multiprocessing.managers import SharedMemoryManager

import structlog
from typing_extensions import Any, Generic, Iterable, Iterator, override

from lenskit.logging.tasks import Task
from lenskit.logging.worker import WorkerContext, WorkerLogConfig

from . import worker
from .config import ParallelConfig, ensure_parallel_init, get_parallel_config, initialize
from .invoker import A, InvokeOp, M, ModelOpInvoker, R
from .serialize import shm_serialize

_log = structlog.stdlib.get_logger(__name__)


def multiprocess_executor(
    n_jobs: int | None = None, sp_config: ParallelConfig | None = None
) -> ProcessPoolExecutor:
    """
    Construct a :class:`ProcessPoolExecutor` configured for LensKit work.
    """
    ensure_parallel_init()
    cfg = get_parallel_config()
    if sp_config is None:
        sp_config = ParallelConfig(1, 1, cfg.child_threads, 1)
    if n_jobs is None:
        n_jobs = cfg.processes

    ctx = LenskitMPContext(sp_config)
    return ProcessPoolExecutor(n_jobs, ctx)


class ProcessPoolOpInvoker(ModelOpInvoker[A, R], Generic[M, A, R]):
    manager: SharedMemoryManager
    pool: ProcessPoolExecutor

    def __init__(
        self,
        model: M,
        func: InvokeOp[M, A, R],
        n_jobs: int,
        worker_parallel: ParallelConfig | None = None,
    ):
        log = _log.bind(n_jobs=n_jobs)
        log.debug("persisting function")
        if worker_parallel is None:
            worker_parallel = ParallelConfig(1, 1, get_parallel_config().child_threads, 1)
        ctx = LenskitMPContext(worker_parallel)

        log.debug("initializing shared memory")
        self.manager = SharedMemoryManager()
        self.manager.start()

        try:
            job = worker.WorkerData(func, model)
            job = shm_serialize(job, self.manager)
            log.info("setting up process pool")
            self.pool = ProcessPoolExecutor(n_jobs, ctx, worker.initalize, (job,))
        except Exception as e:
            self.manager.shutdown()
            raise e

    def map(self, tasks: Iterable[A]) -> Iterator[R]:
        return self.pool.map(worker.worker, self._task_iter(tasks))

    def _task_iter(self, tasks: Iterable[A]):
        """
        Yield the tasks, recording each as dispatched before it is yielded.
        """
        for task in tasks:
            yield task

    def shutdown(self):
        _log.debug("shutting down process pool")
        self.pool.shutdown()
        self.manager.shutdown()
        _log.debug("process pool shut down")


class LenskitProcess(SpawnProcess):
    "LensKit worker process implementation."

    def __init__(
        self, logging: WorkerLogConfig, parallel: ParallelConfig, *args: Any, **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        # save the log config to pass to the process
        self._log_config = logging
        self._parallel_config = parallel

    @override
    def run(self):
        with WorkerContext(self._log_config) as ctx:
            initialize(self._parallel_config)
            log = _log.bind(pid=self.pid, pname=self.name)
            log.info("multiprocessing worker started")
            task = None
            try:
                with Task("worker process", subprocess=True) as task:
                    ctx.send_task(task)
                    super().run()
            finally:
                if task is not None:
                    ctx.send_task(task)


class LenskitMPContext(SpawnContext):
    "LensKit multiprocessing context."

    _log_config: WorkerLogConfig
    _parallel_config: ParallelConfig

    def __init__(self, parallel: ParallelConfig):
        self._log_config = WorkerLogConfig.current()
        self._parallel_config = parallel

    def Process(self, *args: Any, **kwargs: Any) -> SpawnProcess:
        return LenskitProcess(self._log_config, self._parallel_config, *args, **kwargs)



import logging
from datetime import datetime

import numpy as np
import pandas as pd
from typing_extensions import override

from lenskit.data import Dataset, ItemList
from lenskit.pipeline import Component, Trainable

_log = logging.getLogger(__name__)


class PopScorer(Component, Trainable):
    """
    Score items by their popularity.  Use with :py:class:`TopN` to get a
    most-popular-items recommender.

    Args:
        score_type:
            The method for computing popularity scores.  Can be one of the following:

            - ``'quantile'`` (the default)
            - ``'rank'``
            - ``'count'``

    Attributes:
        item_pop_:
            Item popularity scores.
    """

    score_method: str

    item_scores_: pd.Series

    def __init__(self, score_method: str = "quantile"):
        self.score_method = score_method

    @property
    def is_trained(self) -> bool:
        return hasattr(self, "item_scores_")

    @override
    def train(self, data: Dataset):
        _log.info("counting item popularity")
        stats = data.item_stats()
        scores = stats["count"]
        self.item_scores_ = self._train_internal(scores)

        return self

    def _train_internal(self, scores: pd.Series):
        if self.score_method == "rank":
            _log.info("ranking %d items", len(scores))
            scores = scores.rank().sort_index()
        elif self.score_method == "quantile":
            _log.info("computing quantiles for %d items", len(scores))
            cmass = scores.sort_values()
            cmass = cmass.cumsum()
            cdens = cmass / scores.sum()
            scores = cdens.sort_index()
        elif self.score_method == "count":
            _log.info("scoring items with their rating counts")
            scores = scores.sort_index()
        else:
            raise ValueError("invalid scoring method " + repr(self.score_method))

        return scores

    def __call__(self, items: ItemList) -> ItemList:
        scores = self.item_scores_.reindex(items.ids())
        return ItemList(items, scores=scores)

    def __str__(self):
        return "PopScore({})".format(self.score_method)


class TimeBoundedPopScore(PopScorer):
    """
    Score items by their time-bounded popularity, i.e., the popularity in the
    most recent `time_window` period.  Use with :py:class:`TopN` to get a
    most-popular-recent-items recommender.

    Args:
        time_window(datetime.timedelta):
            The time window for computing popularity scores.
        score_type(str):
            The method for computing popularity scores.  Can be one of the following:

            - ``'quantile'`` (the default)
            - ``'rank'``
            - ``'count'``

    Attributes:
        item_scores_(pandas.Series):
            Time-bounded item popularity scores.
    """

    def __init__(self, cutoff: datetime, score_method="quantile"):
        super().__init__(score_method)

        self.cutoff = cutoff
        self.score_method = score_method

    @override
    def train(self, data: Dataset, **kwargs):
        _log.info("counting time-bounded item popularity")

        log = data.interaction_log("numpy")

        item_scores = None
        if log.timestamps is None:
            _log.warning("no timestamps in interaction log; falling back to PopScorer")
            item_scores = super().train(data, **kwargs).item_scores_
        else:
            counts = np.zeros(data.item_count, dtype=np.int32)
            start_timestamp = self.cutoff.timestamp()
            item_nums = log.item_nums[log.timestamps > start_timestamp]
            np.add.at(counts, item_nums, 1)

            item_scores = super()._train_internal(
                pd.Series(counts, index=data.items.index),
                **kwargs,
            )

        self.item_scores_ = item_scores

        return self

    @override
    def __str__(self):
        return "TimeBoundedPopScore({}, {})".format(self.cutoff, self.score_method)



from ._base import Progress
from ._dispatch import item_progress, set_progress_impl
from ._handles import item_progress_handle, pbh_update

__all__ = [
    "Progress",
    "set_progress_impl",
    "item_progress",
    "item_progress_handle",
    "pbh_update",
]
